import torch
import time

import numpy as np
from .torchncon import ncon
from .orthonormalize import left_orthonormalize
from .fixgaugeqr import fixgauge_qr
from .dominanteigensolver import dominant_eigensolver

def symmetrize(A):
    Asymm = A + A.permute(2, 1, 0)
    return Asymm / Asymm.norm()

def H_F(AL, T, x):
    chi, d = AL.shape[0], AL.shape[1]
    x = x.view(chi, d, chi).to(T.device)
    y = ncon([x, T, AL, AL], ([2, 1, 3], [4, 1, 5, -2], [2, 4, -1], [3, 5, -3]))
    return y.contiguous().view(-1).to('cpu')


def H_AC(F, T, x):
    chi, d = F.shape[0], F.shape[1]
    x = x.view(chi, d, chi).to(T.device)
    y = ncon([x, T, F, F], ([2, 1, 3], [1, 4, -2, 5], [2, 4, -1], [3, 5, -3]))
    return y.contiguous().view(-1).to('cpu')


def H_C(F, x):
    chi, d = F.shape[0], F.shape[1]
    x = x.view(chi, chi).to(F.device)
    y = ncon([x, F, F], ([1, 2], [1, 3, -1], [2, 3, -2]))
    return y.contiguous().view(-1).to('cpu')


def vumps_run(T, chi, epsilon, maxiter=100, Cinit=None, ALinit=None, ACinit=None, Finit=None):
    d = T.shape[0]
    dtype, device = T.dtype, T.device

    if Cinit is None:
        F = torch.rand(chi * d * chi, dtype=dtype, device=device)
        F = symmetrize(F.view(chi, d, chi))

        A = torch.rand(chi, d, chi, dtype=dtype, device=device)
        A = symmetrize(A)

        AL, C, _ = left_orthonormalize(A, torch.eye(chi, chi, dtype=dtype, device=device), epsilon / 10)

        AC = C @ A.view(chi, d * chi)
        AC = AC.view(chi * d, chi) @ C.t()
        AC = symmetrize(AC.view(chi, d, chi))
    else:
        F = Finit.detach()
        AL = ALinit.detach()
        AC = ACinit.detach()
        C = Cinit.detach()
    # start = time.time()
    # rtn_history = []

    for it in range(maxiter):
        print ('it', it, maxiter, epsilon)
        lambF, F = dominant_eigensolver(lambda x: H_F(AL, T, x), F.view(-1).to('cpu'), tol=epsilon / 10)
        print ('got F')
        F = symmetrize(F.view(chi, d, chi)).to(device)

        # compute error
        MAC = H_AC(F, T, (AL.view(chi * d, chi) @ C).view(-1)).view(chi, d, chi).to(device)
        MAC = MAC - ncon([MAC, AL, AL], ([1, 2, -3], [1, 2, 3], [-1, -2, 3]))
        diff = MAC.norm()

        lambAC, AC = dominant_eigensolver(lambda x: H_AC(F, T, x), AC.view(-1).to('cpu'), tol=epsilon / 10)

        print ('got AC')
        AC = symmetrize(AC.view(chi, d, chi).to(device))

        lambC, C = dominant_eigensolver(lambda x: H_C(F, x), C.view(-1).to('cpu'), tol=epsilon / 10)

        print ('got C')
        C = C.view(chi, chi).to(device)
        C = (C + C.t()) / 2
        C = C / C.norm()

        QAC, RAC = fixgauge_qr(AC.view(chi * d, chi))
        QC, RC = fixgauge_qr(C)
        print ('qr done')
        AL = (QAC @ QC.t()).view(chi, d, chi)
        # diff = torch.dist(RAC, RC)

        # print(it, diff.item(), torch.log(lambF).item(), torch.log(lambAC / lambC).item())
        # if it % 5 == 0:
        #     rtn_history.append([time.time() - start, torch.log(lambAC / lambC).item()])
        if (diff < epsilon):
            break

    if diff >= epsilon:
        print('WARNING: vumps reaching maxiter, diff=', diff.item())

    return torch.log(lambAC / lambC), AC, F, C, AL


def vumps_calc(T, AC, F, C):
    return torch.dot(H_F(AC, T, F).to(F.device), F.view(-1)) \
           / torch.dot(AC.view(-1), AC.view(-1)) \
           / torch.dot(H_C(F, C).to(F.device), C.view(-1))

